/****************************************************************************/
// Eclipse SUMO, Simulation of Urban MObility; see https://eclipse.org/sumo
// Copyright (C) 2001-2022 German Aerospace Center (DLR) and others.
// This program and the accompanying materials are made available under the
// terms of the Eclipse Public License 2.0 which is available at
// https://www.eclipse.org/legal/epl-2.0/
// This Source Code may also be made available under the following Secondary
// Licenses when the conditions for such availability set forth in the Eclipse
// Public License 2.0 are satisfied: GNU General Public License, version 2
// or later which is available at
// https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html
// SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
/****************************************************************************/
/// @file    Select_cursor.cpp
/// @author  Angelo Banse
/// @date    2021-01-22
///
// Generated by reswrap version 4.0.0
// created by reswrap from file Select_cursor.gif
/****************************************************************************/

const unsigned char Select_cursor[] = {
    0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x20, 0x00, 0x20, 0x00, 0xa1, 0x02, 0x00, 0x00, 0x00, 0x5a,
    0x00, 0x64, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x21, 0xf9, 0x04, 0x01, 0x0a, 0x00, 0x02,
    0x00, 0x2c, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x20, 0x00, 0x00, 0x02, 0x4c, 0x94, 0x8f, 0xa9,
    0x9b, 0xc0, 0x0f, 0x11, 0x70, 0xb1, 0x2a, 0x10, 0xa8, 0xdd, 0x38, 0x6f, 0x1e, 0x84, 0xda, 0xf7,
    0x74, 0x22, 0x09, 0x99, 0x27, 0xba, 0xa8, 0x2b, 0x2b, 0x85, 0xf2, 0x0b, 0x0b, 0xae, 0x3c, 0xb2,
    0x37, 0x5e, 0xdb, 0xf3, 0xef, 0x81, 0xdd, 0x26, 0xc4, 0x9a, 0xa9, 0x93, 0x13, 0x9e, 0x90, 0x3d,
    0x03, 0x86, 0xc2, 0x6c, 0x4e, 0x62, 0xc1, 0x26, 0x35, 0x69, 0x9d, 0x5a, 0xb7, 0xdc, 0xae, 0xf7,
    0x0b, 0x0e, 0x8b, 0xc7, 0xe4, 0xb2, 0xf9, 0x5b, 0x00, 0x00, 0x3b
};


/****************************************************************************/
